import json

# JSON Utils
JSON_CONTENT_TYPES = ('application/json', 'text/json', 'text/javascript')


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()

        return super(DateTimeJSONEncoder, self).default(obj)


class JSONDecoder(json.JSONDecoder):
    pass



from typing import Optional, Dict
import time
from random import random
import hashlib

from rivr.middleware import Middleware
from rivr.request import Request
from rivr.response import Response


class BaseSession(object):
    def __init__(self, session_key: Optional[str] = None):
        self.modified = False
        self.session_key = session_key
        self.data: Dict[str, str] = {}

        if self.session_key:
            self.get_session()

    def get_session(self):
        raise NotImplementedError

    def save(self):
        raise NotImplementedError

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __getitem__(self, key: str) -> str:
        return self.data[key]

    def __setitem__(self, key: str, value: str):
        self.modified = True
        self.data[key] = value

    def __delitem__(self, key: str):
        self.modified = True
        del self.data[key]

    def clear(self):
        self.modified = True
        self.data = {}

    def generate_key(self):
        self.modified = True
        hsh = hashlib.sha1((str(time.time()) + str(random())).encode('utf-8'))
        self.session_key = hsh.hexdigest()


class MemorySession(BaseSession):
    def __init__(self, store, *args, **kwargs):
        self.store = store
        super(MemorySession, self).__init__(*args, **kwargs)

    def get_session(self):
        if self.session_key in self.store.sessions:
            self.data = self.store.sessions[self.session_key]

    def save(self):
        self.store.sessions[self.session_key] = self.data


class MemorySessionStore(object):
    def __init__(self):
        self.sessions = {}

    def __call__(self, *args, **kwargs):
        return MemorySession(self, *args, **kwargs)


class SessionMiddleware(Middleware):
    cookie_name = 'sessionid'
    cookie_secure = False
    cookie_path = '/'
    cookie_domain = None

    session_store = None

    def process_request(self, request: Request) -> Optional[Response]:
        if self.session_store is None:
            raise Exception(
                "SessionMiddleware is improperly configured."
                "Session store is not defined"
            )

        session_key = request.COOKIES.get(self.cookie_name, None)
        request.session = self.session_store(session_key)

    def process_response(self, request: Request, response: Response) -> Response:
        session = getattr(request, 'session')
        if session and session.modified:
            # Save the session data and refresh the client cookie.
            session.save()

            if not session.session_key:
                session.generate_key()

            response.set_cookie(
                self.cookie_name,
                session.session_key,
                path=self.cookie_path,
                domain=self.cookie_domain,
                secure=self.cookie_secure,
            )

        return response



from typing import Callable
from wsgiref.simple_server import WSGIServer, WSGIRequestHandler

from rivr.wsgi import WSGIHandler
from rivr.middleware.debug import DebugMiddleware
from rivr.request import Request
from rivr.response import Response

__all__ = ['serve']


def serve(
    handler: Callable[[Request], Response],
    host: str = 'localhost',
    port: int = 8080,
    debug: bool = True,
):
    """
    Starts a developent server on the local machine. By default, the
    server runs on port 8080 on localhost. You can pass in a different
    hostname and/or IP using the keyword arguments.
    """

    if debug:
        handler = DebugMiddleware.wrap(handler)

    httpd = WSGIServer((host, port), WSGIRequestHandler)
    httpd.set_app(WSGIHandler(handler))
    print("Development server is running at http://%s:%s/" % (host, port))
    httpd.serve_forever()



from typing import Callable, Dict, Optional
from rivr.request import Request
from rivr.response import Response

__all__ = ['TestClient']


class TestClient(object):
    def __init__(self, handler: Callable[[Request], Response]):
        self.handler = handler

    def http(
        self,
        method: str,
        path: str,
        data=None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Response:
        if method == 'GET':
            get = data
        else:
            get = {}

        request = Request(path, method, get, data, headers)

        return self.handler(request)

    def get(self, *args, **kwargs) -> Response:
        return self.http('GET', *args, **kwargs)

    def put(self, *args, **kwargs) -> Response:
        return self.http('PUT', *args, **kwargs)

    def post(self, *args, **kwargs) -> Response:
        return self.http('POST', *args, **kwargs)

    def delete(self, *args, **kwargs) -> Response:
        return self.http('DELETE', *args, **kwargs)


TestClient.__test__ = False  # type: ignore



import os
import posixpath
from urllib.parse import unquote
from functools import reduce
import mimetypes
import stat
import re
from email.utils import formatdate, parsedate_tz, mktime_tz

from rivr.views.base import View
from rivr.request import Request
from rivr.response import Response, ResponseRedirect, ResponseNotModified, Http404


DEFAULT_DIRECTORY_INDEX_TEMPLATE = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
    "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
<head>
    <meta http-equiv="Content-type" content="text/html; charset=utf-8" />
    <meta http-equiv="Content-Language" content="en-us" />
    <meta name="robots" content="NONE,NOARCHIVE" />
    <title>Index of {directory}</title>
</head>

<body>
    <h1>Index of {directory}</h1>
    <ul>
        {files}
    </ul>
</body>
"""


class StaticView(View):
    """
    Usage in the router:

        (r'(?P<path>.*)$', StaticView.as_view(document_root='.',
                                              show_indexes=True))
    """

    document_root = None
    show_indexes = False
    use_request_path = False
    index = None

    def was_modified_since(self, mtime=0, size=0):
        header = self.request.headers.get('HTTP_IF_MODIFIED_SINCE')
        if header is None:
            return True

        matches = re.match(r'^([^;]+)(; length=([0-9]+))?$', header, re.IGNORECASE)

        header_mtime = mktime_tz(parsedate_tz(matches.group(1)))
        header_len = matches.group(3)

        if header_len and int(header_len) != size:
            return True

        if mtime > header_mtime:
            return True

        return False

    def directory_index(self, request: Request, path: str, fullpath: str) -> Response:
        if not self.show_indexes:
            raise Http404('Directory indexes are not allowed here.')

        directories = []
        files = []
        for f in os.listdir(fullpath):
            if not f.startswith('.'):
                if os.path.isdir(os.path.join(fullpath, f)):
                    directories.append(f + '/')
                else:
                    files.append(f)

        if path != '':
            directories.append('../')

        directories.sort()
        files.sort()

        def render(accumulator, path):
            return accumulator + '<li><a href="{0}">{0}</a></li>\n'.format(path)

        context = {
            'directory': path + '/',
            'files': reduce(render, directories + files, ''),
        }

        return Response(DEFAULT_DIRECTORY_INDEX_TEMPLATE.format(**context))

    def get(self, request: Request, path: str = '') -> Response:
        if not self.document_root:
            raise Exception('Improperly configured view')

        if self.use_request_path:
            path = request.path

        # Clean up given path to only allow serving files below document_root.
        path = posixpath.normpath(unquote(path))
        path = path.lstrip('/')
        newpath = ''

        for part in path.split('/'):
            if not part:
                # Strip empty path components.
                continue

            drive, part = os.path.splitdrive(part)
            head, part = os.path.split(part)

            if part in (os.curdir, os.pardir):
                # Strip '.' and '..' in path.
                continue

            newpath = os.path.join(newpath, part).replace('\\', '/')

        if newpath and path != newpath:
            return ResponseRedirect(newpath)

        fullpath = os.path.join(self.document_root, newpath)

        if not os.path.exists(fullpath):
            raise Http404('"%s" does not exist.' % newpath)

        if os.path.isdir(fullpath):
            if self.index:
                index = os.path.join(fullpath, self.index)

                if os.path.exists(index) and not os.path.isdir(index):
                    return self.file(index)

            return self.directory_index(request, newpath, fullpath)

        return self.file(fullpath)

    def file(self, fullpath: str) -> Response:
        statobj = os.stat(fullpath)

        if not self.was_modified_since(statobj[stat.ST_MTIME], statobj[stat.ST_SIZE]):
            return ResponseNotModified()

        mimetype = mimetypes.guess_type(fullpath)[0] or 'application/octet-stream'

        with open(fullpath, 'rb') as fp:
            contents = fp.read()

        response = Response(contents, content_type=mimetype)
        response.headers['Last-Modified'] = '%s GMT' % (
            formatdate(statobj[stat.ST_MTIME])[:25]
        )
        response.headers['Content-Length'] = str(len(contents))

        return response



from typing import Callable
from rivr.response import Response


def import_module(name: str) -> Callable[..., Response]:
    if callable(name):
        return name

    if not isinstance(name, str):
        raise ImportError(
            'rivr.import.import_module cannot import %s because it is not a string.'
            % name
        )

    module = __import__(name)
    for i in name.split('.')[1:]:
        module = getattr(module, i)

    return module



from typing import Optional, Callable
from rivr.request import Request
from rivr.response import Response

__all__ = ['Middleware']


class Middleware(object):
    @classmethod
    def wrap(cls, view: Callable[..., Response], *initargs, **initkwargs):
        """
        The wrap method allows you to wrap a view calling the middleware's
        process_request and process_response before and after the view.

        If the view raises an exception, the process_exception method will be
        called.

        Example::

            view = Middleware.wrap(view)
            response = view(request)

        """

        def func(*args, **kwargs) -> Response:
            return cls(*initargs, **initkwargs).dispatch(view, *args, **kwargs)

        return func

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def dispatch(
        self, view: Callable[..., Response], request: Request, *args, **kwargs
    ) -> Response:
        response = self.process_request(request)

        if response is None:
            try:
                response = view(request, *args, **kwargs)
            except Exception as e:
                process_exception: Optional[
                    Callable[[Request, Exception], Optional[Response]]
                ] = getattr(self, 'process_exception')
                if process_exception:
                    response = process_exception(request, e)
                    if not response:
                        raise
                else:
                    raise

        return self.process_response(request, response)

    def process_request(self, request: Request) -> Optional[Response]:
        """
        This method is called before the view on each request. This method
        should either return a response or None. If it returns a response, then
        the middleware will not call the view. If it returns None, then we will
        call the view.
        """
        return None

    def process_response(self, request: Request, response: Response) -> Response:
        """
        This method will take the response, either from process_request or the
        view. This method will always be called for each request unless there
        is an exception.

        This method must return a response, this can either be the response
        passed to it, or a completely new response.
        """
        return response



from typing import Callable, List, Optional
from email.utils import formatdate, parsedate
from datetime import datetime
from calendar import timegm

from rivr.request import Request
from rivr.response import (
    Http404,
    Response,
    ResponseNotAllowed,
    ResponseRedirect,
    ResponsePermanentRedirect,
    RESTResponse,
    ResponseNoContent,
    ResponseNotModified,
)

__all__ = ['View', 'RedirectView']


class View(object):
    http_method_names = ['get', 'post', 'put', 'delete', 'head', 'options', 'trace']

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def as_view(cls, **initkwargs) -> Callable[..., Response]:
        """
        This method will return a callable which will generate a new instance
        of the class passing it the kwargs passed to it.

        Usage::

            view = View.as_view()
            response = view(request)

        """

        def view(*args, **kwargs) -> Response:
            return cls(**initkwargs).dispatch(*args, **kwargs)

        return view

    def get_handler(self, request: Request) -> Callable[..., Response]:
        if request.method.lower() in self.http_method_names:
            return getattr(self, request.method.lower(), self.http_method_not_allowed)
        return self.http_method_not_allowed

    def dispatch(self, request: Request, *args, **kwargs) -> Response:
        self.request = request
        self.args = args
        self.kwargs = kwargs

        return self.get_handler(request)(request, *args, **kwargs)

    def http_method_not_allowed(self, request: Request, *args, **kwargs) -> Response:
        return ResponseNotAllowed(self.get_allowed_methods())

    def get_allowed_methods(self) -> List[str]:
        allowed_methods = [m for m in self.http_method_names if hasattr(self, m)]
        if not hasattr(self, 'get'):
            allowed_methods.remove('head')

        return allowed_methods

    def options(self, request: Request, *args, **kwargs) -> Response:
        allowed_methods = [m.upper() for m in self.get_allowed_methods()]
        response = ResponseNoContent()
        response.headers['Allow'] = ','.join(allowed_methods)
        return response

    def head(self, request: Request, *args, **kwargs) -> Response:
        get = getattr(self, 'get', None)
        if get:
            response = get(request, *args, **kwargs)
            response.content = ''
            return response

        return self.http_method_not_allowed(request, *args, **kwargs)


class RedirectView(View):
    """
    The redirect view will redirect on any GET requests.
    """

    permanent = True
    url: Optional[str] = None
    query_string = False

    def get_redirect_url(self, **kwargs) -> Optional[str]:
        """
        Return the URL we should redirect to
        """

        if self.url:
            args = self.request.META["QUERY_STRING"]
            if args and self.query_string:
                url = "%s?%s" % (self.url, args)
            else:
                url = self.url
            return url % kwargs
        return None

    def get(self, request: Request, *args, **kwargs) -> Response:
        url = self.get_redirect_url(**kwargs)
        if url:
            if self.permanent:
                return ResponsePermanentRedirect(url)
            return ResponseRedirect(url)
        raise Http404("Redirect URL not found.")



from typing import Optional
from base64 import b64decode

from rivr.request import Request
from rivr.response import Response
from rivr.middleware.base import Middleware


class ResponseAuthorizationRequired(Response):
    status_code = 401

    def __init__(self, *args, **kwargs):
        super(ResponseAuthorizationRequired, self).__init__(*args, **kwargs)
        self.headers['WWW-Authenticate'] = 'Basic realm="Secure Area"'


class User(object):
    def __init__(self, username: str = ''):
        self.username = username

    def is_anonymous(self) -> bool:
        return False

    def is_authenticated(self) -> bool:
        return True


class AnnonymousUser(object):
    username = ''

    def is_anonymous(self) -> bool:
        return True

    def is_authenticated(self) -> bool:
        return False


class AuthMiddleware(Middleware):
    needs_auth = True

    def process_request(self, request: Request) -> Optional[Response]:
        user = self.check_login(request)
        setattr(request, 'user', user)

        if not user.is_authenticated() and self.needs_auth:
            return ResponseAuthorizationRequired()

        return None

    def check_login(self, request: Request):
        authorization = request.META.get('HTTP_AUTHORIZATION', None)

        if not authorization:
            return AnnonymousUser()

        try:
            method, key = authorization.split()
        except ValueError:
            return AnnonymousUser()

        if method != 'Basic':
            return AnnonymousUser()

        try:
            username, password = b64decode(key).decode('utf-8').split(':')
        except ValueError:
            return AnnonymousUser()

        if self.check_password(username, password):
            return User(username)

        return AnnonymousUser()

    def check_password(self, username: str, password: str) -> bool:
        return False



from typing import Optional, Dict, Any, Callable, Iterable, IO
import sys
import logging
from urllib.parse import parse_qsl

from rivr.request import parse_cookie, Request
from rivr.response import Response, ResponseNotFound, Http404
from rivr.utils import JSON_CONTENT_TYPES, JSONDecoder

logger = logging.getLogger('rivr.request')

STATUS_CODES = {
    100: 'CONTINUE',
    101: 'SWITCHING PROTOCOLS',
    200: 'OK',
    201: 'CREATED',
    202: 'ACCEPTED',
    203: 'NON-AUTHORITATIVE INFORMATION',
    204: 'NO CONTENT',
    205: 'RESET CONTENT',
    206: 'PARTIAL CONTENT',
    300: 'MULTIPLE CHOICES',
    301: 'MOVED PERMANENTLY',
    302: 'FOUND',
    303: 'SEE OTHER',
    304: 'NOT MODIFIED',
    305: 'USE PROXY',
    306: 'RESERVED',
    307: 'TEMPORARY REDIRECT',
    400: 'BAD REQUEST',
    401: 'UNAUTHORIZED',
    402: 'PAYMENT REQUIRED',
    403: 'FORBIDDEN',
    404: 'NOT FOUND',
    405: 'METHOD NOT ALLOWED',
    406: 'NOT ACCEPTABLE',
    407: 'PROXY AUTHENTICATION REQUIRED',
    408: 'REQUEST TIMEOUT',
    409: 'CONFLICT',
    410: 'GONE',
    411: 'LENGTH REQUIRED',
    412: 'PRECONDITION FAILED',
    413: 'REQUEST ENTITY TOO LARGE',
    414: 'REQUEST-URI TOO LONG',
    415: 'UNSUPPORTED MEDIA TYPE',
    416: 'REQUESTED RANGE NOT SATISFIABLE',
    417: 'EXPECTATION FAILED',
    500: 'INTERNAL SERVER ERROR',
    501: 'NOT IMPLEMENTED',
    502: 'BAD GATEWAY',
    503: 'SERVICE UNAVAILABLE',
    504: 'GATEWAY TIMEOUT',
    505: 'HTTP VERSION NOT SUPPORTED',
}


class WSGIRequest(object):
    """
    https://wsgi.readthedocs.io/en/latest/definitions.html
    """

    def __init__(self, environ: Dict[str, Any]):
        self.environ = environ

        self.method: str = environ['REQUEST_METHOD']
        self.path: str = environ['PATH_INFO']
        self.META = environ

    @property
    def headers(self) -> Dict[str, Any]:
        if not hasattr(self, '_headers'):
            headers = {}

            for key in self.META:
                if key.startswith('HTTP_'):
                    headers[key[5:]] = self.META[key]

            self._headers = headers
        return self._headers

    @property
    def is_secure(self) -> bool:
        """
        Returns True if connection was made over HTTPS/TLS.
        """

        return self.scheme == 'https'

    @property
    def scheme(self) -> str:
        """
        Scheme used for the request. For example, `https`.
        """

        return self.environ['wsgi.url_scheme']

    @property
    def host(self) -> str:
        """
        Hostname used for the request. For example, `rivr.com`.
        """

        host = self.environ.get('HTTP_HOST', None)

        if host is None:
            host = self.environ.get('SERVER_NAME')

        return host

    @property
    def port(self) -> int:
        """
        Port used for the connection.
        """

        return int(self.environ['SERVER_PORT'])

    @property
    def url(self):
        """
        Hostname used for the request. For example, `https://rivr.com/about`.
        """

        scheme = self.scheme
        url = scheme + '://' + self.host
        port = self.port
        if (scheme == 'http' and port != 80) or (scheme == 'https' and port != 443):
            url += ':' + str(port)

        return url + self.path

    @property
    def content_length(self) -> int:
        """
        Returns the length of the request's body.
        """

        try:
            content_length = int(self.environ.get('CONTENT_LENGTH', 0))
        except (ValueError, TypeError):
            content_length = 0

        return content_length

    @property
    def content_type(self) -> Optional[str]:
        return self.environ.get('CONTENT_TYPE', None)

    @property
    def body(self) -> IO[bytes]:
        """
        Request body (file descriptor).
        """

        return self.environ['wsgi.input']

    # @property
    def get_get(self):
        if not hasattr(self, '_get'):
            query_string = self.environ.get('QUERY_STRING', '')
            self._get = dict((k, v) for k, v in parse_qsl(query_string, True))
        return self._get

    GET = property(get_get)

    # Attributes

    @property
    def attributes(self):
        """
        A request body, deserialized as a dictionary.

        This will automatically deserialize form or JSON encoded request
        bodies.
        """

        if not hasattr(self, '_attributes'):
            if self.content_length > 0:
                content = self.body.read(self.content_length)
                content_type = self.content_type.split(';')[0]

                if content_type in JSON_CONTENT_TYPES:
                    content = content.decode('utf-8')
                    self._attributes = JSONDecoder().decode(content)
                elif content_type == 'application/x-www-form-urlencoded':
                    content = content.decode('utf-8')
                    data = parse_qsl(content, True)
                    self._attributes = dict((k, v) for k, v in data)
            else:
                self._attributes = {}

        return self._attributes

    POST = attributes  # Deprecated

    @property
    def cookies(self):
        """
        Dictionary returning all of the cookies sent with the request.
        """

        if not hasattr(self, '_cookies'):
            self._cookies = parse_cookie(self.headers.get('COOKIE', ''))
        return self._cookies

    COOKIES = cookies  # legacy


class WSGIHandler(object):
    request_class = WSGIRequest

    def __init__(self, view: Callable[..., Response]):
        self.view = view

    def __call__(
        self, environ: Dict[str, Any], start_response: Callable[[str, Any], Any]
    ) -> Iterable[bytes]:
        request = self.request_class(environ)

        try:
            response = self.view(request)
            if not response:
                raise Exception("View did not return a response.")
        except Http404:
            response = ResponseNotFound('Page not found')
        except Exception:
            logger.error(
                'Internal Server Error: {}'.format(request.path),
                exc_info=sys.exc_info(),
                extra={'status_code': 500, 'request': request},
            )

            response = Response('Internal server error', status=500)

        try:
            status_text = STATUS_CODES[response.status_code]
        except KeyError:
            status_text = 'UNKNOWN STATUS CODE'

        status = '%s %s' % (response.status_code, status_text)

        start_response(status, response.headers_items())

        response_content = response.content

        if isinstance(response_content, str):
            response_content = response_content.encode('utf-8')
        elif sys.version_info[0] < 3 and isinstance(response.content, unicode):
            response_content = response_content.encode('utf-8')

        return [response_content]



from typing import Dict, Optional
from http.cookies import SimpleCookie, CookieError

__all__ = ['Request']


def parse_cookie(cookie):
    if cookie == '':
        return {}

    try:
        c = SimpleCookie()
        c.load(cookie)
    except CookieError:
        return {}

    cookiedict = {}
    for key in c.keys():
        cookiedict[key] = c.get(key).value

    return cookiedict


class Request(object):
    """
    A request is an object which represents a HTTP request. You wouldn't
    normally create a request yourself but instead be passed a request. Each
    view gets passed the clients request.
    """

    def __init__(
        self,
        path: str = '/',
        method: str = 'GET',
        get: Optional[Dict[str, str]] = None,
        attributes=None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.path = path
        self.method = method
        self.GET = get or {}
        self.attributes = attributes or {}
        self.headers = headers or {}
        self.META: Dict[str, str] = {}

    @property
    def cookies(self):
        return parse_cookie(self.headers.get('COOKIE', ''))

    COOKIES = cookies  # legacy



from typing import Optional, Callable, List
from rivr.middleware.base import Middleware
from rivr.request import Request
from rivr.response import Response, ResponseNotFound, Http404


class MiddlewareController(Middleware):
    """
    The middleware controller allows you to wrap a view in multiple middleware.

    Example usage::

        view = MiddlewareController.wrap(view,
            FirstMiddleware(),
            SecondMiddleware()
        )

        response = view(request)
    """

    def __init__(self, *middleware):
        self.request_middleware: List[Callable[[Request], Optional[Response]]] = []
        self.response_middleware: List[Callable[[Request, Response], Response]] = []
        self.exception_middleware: List[
            Callable[[Request, Exception], Optional[Response]]
        ] = []

        for mw_instance in middleware:
            self.append(mw_instance)

    def append(self, middleware: Middleware):
        if hasattr(middleware, 'process_request'):
            self.request_middleware.append(middleware.process_request)

        if hasattr(middleware, 'process_response'):
            self.response_middleware.insert(0, middleware.process_response)

        process_exception = getattr(middleware, 'process_exception', None)
        if process_exception:
            self.exception_middleware.insert(0, process_exception)

    def process_request(self, request: Request) -> Optional[Response]:
        for request_mw in self.request_middleware:
            response = request_mw(request)
            if response:
                return response

        return None

    def process_response(self, request: Request, response: Response) -> Response:
        for response_mw in self.response_middleware:
            response = response_mw(request, response)

        return response

    def process_exception(
        self, request: Request, exception: Exception
    ) -> Optional[Response]:
        for exception_mw in self.exception_middleware:
            response = exception_mw(request, exception)
            if response:
                return response
        else:
            raise


class ErrorWrapper(object):
    def __init__(
        self,
        app: Callable[..., Response],
        custom_404: Optional[Callable[[Request, Exception], Response]] = None,
        custom_500: Optional[Callable[[Request, Exception], Response]] = None,
    ):
        self.app = app
        self.error_404 = custom_404 or self.default_error_404
        self.error_500 = custom_500 or self.default_error_500

    def __call__(self, request: Request, *args, **kwargs) -> Response:
        try:
            response = self.app(request, *args, **kwargs)
        except Http404 as e:
            response = self.error_404(request, e)
        except Exception as e:
            response = self.error_500(request, e)

        if not response:
            response = self.error_404(request, Http404())

        return response

    def default_error_404(self, request: Request, e: Exception) -> Response:
        return ResponseNotFound("404: %s" % e)

    def default_error_500(self, request: Request, e: Exception) -> Response:
        return Response("A 500 has occured", status=500)



VERSION = '0.8.1'

from rivr.response import Response, Http404
from rivr.server import serve

from rivr.middleware import Middleware, MiddlewareController, ErrorWrapper
from rivr.middleware.debug import DebugMiddleware
from rivr.middleware.auth import AuthMiddleware

from rivr.router import Router, Domain, url, include

from rivr import views



import sys

try:
    from pygments import highlight  # type: ignore
    from pygments.lexers import PythonTracebackLexer  # type: ignore
    from pygments.formatters import HtmlFormatter  # type: ignore

    HAS_PYGMENTS = True
except ImportError:
    HAS_PYGMENTS = False

from rivr import VERSION
from rivr.middleware import Middleware
from rivr.request import Request
from rivr.response import Response, ResponseNotFound, Http404

__all__ = ['DebugMiddleware']

ERROR_CSS = """
    html * { padding:0; margin:0; }
    body * { padding:10px 20px; font-family: 'Helvetica Neue', HelveticaNeue, Arial, Helvetica, sans-serif; color: #555; }
    body * * { padding:0; }
    h1, h2 { font-weight: normal; margin-bottom:.4em; letter-spacing: -1px; }
    h1 { font-size: 27px; }
    h2 { font-size: 22px; }
    header { background: #eeeeee; border-bottom: 1px solid #ddd; }
    header h1 { text-shadow: #fff 2px 2px 0; }
    #content { margin: 5px; }
    pre { border: 1px solid #ddd; margin: 10px 0; background: #eee; width: 720px; padding: 10px; border-radius: 6px; overflow: auto; font-size: 9pt; font-weight: normal; }
    pre:hover { border: 1px solid #bbb; }
    code { font-family: Consolas, Monaco, 'Lucida Console', 'Liberation Mono', 'DejaVu Sans Mono', 'Bitstream Vera Sans Mono', 'Courier New'; }
    footer { width: 720px; margin-top: 25px; font-size: 0.9em; }
"""

ERROR_TEMPLATE = """
<!DOCTYPE HTML>
<html lang="en">
<head>
    <meta charset="utf-8">
    <title>{title}</title>
    <style type="text/css" media="screen">
    {css}
    </style>
</head>
<body>
    <header>
        <h1>{title}</h1>
    </header>

    <div id="content">
        <section>
            <h2>Traceback</h2>
            {traceback}
        </section>

        <section>
            <h2>Request</h2>
            <pre><code>{request}</code></pre>
        </section>

        <footer>
            <p>Powered by rivr {version} on Python {python_version}.</p>
        </footer>
    </div>
</body>
"""


class DebugMiddleware(Middleware):
    def process_404(self, request: Request) -> Response:
        return ResponseNotFound('Page Not Found')

    def process_exception(self, request: Request, e: Exception) -> Response:
        if isinstance(e, Http404):
            return self.process_404(request)

        import traceback

        tb = '\n'.join(traceback.format_exception(*sys.exc_info()))

        if HAS_PYGMENTS:
            formatter = HtmlFormatter(style='borland')
            html_tb = highlight(tb, PythonTracebackLexer(), formatter)
        else:
            html_tb = '<pre><code>%s</pre></code>' % (tb)

        context = {
            'title': str(e),
            'traceback': html_tb,
            'request': repr(request).replace('>', '&gt;').replace('<', '&lt;'),
            'version': VERSION,
            'python_version': '{}.{}.{}-{}-{}'.format(*sys.version_info),
            'css': ERROR_CSS,
        }

        if HAS_PYGMENTS:
            context['css'] += formatter.get_style_defs('.highlight')

        return Response(ERROR_TEMPLATE.format(**context), status=500)

    def process_response(self, request: Request, response: Response) -> Response:
        if not response:
            return self.process_exception(
                request, Exception("View did not return a response.")
            )
        return response



from typing import Optional, List, Union, Iterable, Tuple
from http.cookies import SimpleCookie
import datetime

from rivr.request import Request
from rivr.utils import JSON_CONTENT_TYPES, JSONEncoder


class Http404(Exception):
    pass


class Response(object):
    """
    Response is an object for describing a HTTP response. Every view is
    responsible for either returning a response or raising an exception.
    """

    status_code: int = 200
    """The HTTP status code for the response."""

    def __init__(
        self,
        content: Union[str, bytes] = '',
        status: Optional[int] = None,
        content_type: Optional[str] = 'text/html; charset=utf8',
    ):
        self.content = content

        if status:
            self.status_code = status

        self.headers = {}
        if content_type:
            self.headers['Content-Type'] = content_type

        self.cookies: SimpleCookie = SimpleCookie()

    @property
    def content_type(self) -> str:
        return self.headers['Content-Type']

    def __str__(self) -> str:
        headers = ['%s: %s' % (key, value) for key, value in self.headers.items()]
        if isinstance(self.content, bytes):
            return '\n'.join(headers) + '\n\n' + self.content.decode('utf-8')

        return '\n'.join(headers) + '\n\n' + self.content

    def set_cookie(
        self,
        key: str,
        value: str = '',
        max_age=None,
        expires: Union[datetime.datetime, str] = None,
        path: str = '/',
        domain: Optional[str] = None,
        secure: bool = False,
    ):
        """
        Sets a cookie. The parameters are the same as in the Cookie.Morsel
        object in the Python standard library.
        """

        self.cookies[key] = value

        if max_age is not None:
            self.cookies[key]['max-age'] = max_age

        if expires is not None:
            if isinstance(expires, datetime.datetime):
                expires = expires.strftime('%a, %d %b %Y %H:%M:%S')

            self.cookies[key]['expires'] = expires

        if path is not None:
            self.cookies[key]['path'] = path

        if domain is not None:
            self.cookies[key]['domain'] = domain

        if secure:
            self.cookies[key]['secure'] = True

    def delete_cookie(self, key: str, path: str = '/', domain: Optional[str] = None):
        """
        Deletes the cookie with the given key. Fails silently if the key
        doesn't exist
        """

        self.set_cookie(
            key,
            max_age=0,
            path=path,
            domain=domain,
            expires='Thu, 01-Jan-1970 00:00:00 GMT',
        )

    def headers_items(self) -> Iterable[Tuple[str, str]]:
        headers = [(k, v) for k, v in self.headers.items()]

        for cookie in self.cookies.values():
            headers.append(('Set-Cookie', str(cookie.output(header=''))))

        return headers


class ResponseNoContent(Response):
    """A response that uses the 204 status code to indicate no content."""

    status_code = 204


class ResponseRedirect(Response):
    """
    Acts just like a ResponseRedirect, but uses a 302 status code.
    It takes a URL to redirect the user to.
    """

    status_code = 302

    def __init__(self, redirect_to: str):
        super(ResponseRedirect, self).__init__()
        self.headers['Location'] = str(redirect_to)

    @property
    def url(self) -> str:
        """
        A property that returns the URL for the redirect.
        """
        return self.headers['Location']


class ResponsePermanentRedirect(ResponseRedirect):
    """Acts just like a ResponseRedirect, but uses a 301 status code."""

    status_code = 301


class ResponseNotFound(Response):
    """Acts just like a Response, but uses a 404 status code."""

    status_code = 404


class ResponseNotModified(Response):
    """Acts just like a Response, but uses a 304 status code."""

    status_code = 304


class ResponseNotAllowed(Response):
    """
    A response that uses the 405 status code and takes a list of
    permitted HTTP methods.
    """

    status_code = 405

    def __init__(self, permitted_methods: List[str]):
        super(ResponseNotAllowed, self).__init__()
        self.headers['Allow'] = ', '.join(permitted_methods)


class RESTResponse(Response):
    def __init__(self, request: Request, payload, status: Optional[int] = None):
        content = JSONEncoder().encode(payload)
        content_type = JSON_CONTENT_TYPES[0]

        super(RESTResponse, self).__init__(content, status, content_type)



from typing import Callable, Optional, Tuple, Iterable, Any, Dict
import re

from rivr.importlib import import_module
from rivr.request import Request
from rivr.response import Http404, Response, ResponsePermanentRedirect


class Resolver404(Http404):
    pass


class ViewDoesNotExist(Exception):
    pass


class RegexURL(object):
    def __init__(self, regex: str):
        self.regex = re.compile(regex, re.UNICODE)

    def resolve(
        self, path: str
    ) -> Optional[Tuple[Callable, Iterable[Any], Dict[str, Any]]]:
        match = self.regex.search(path)
        if match:
            return self.match_found(path, match)

        return None

    def match_found(
        self, path: str, match
    ) -> Optional[Tuple[Callable, Iterable[Any], Dict[str, Any]]]:
        raise NotImplementedError


class RegexURLPattern(RegexURL):
    def __init__(
        self,
        regex: str,
        callback: Callable[..., Response],
        kwargs={},
        name: Optional[str] = None,
        prefix: Optional[str] = None,
    ):
        super(RegexURLPattern, self).__init__(regex)

        if callable(callback):
            self._callback = callback
        else:
            self._callback = None
            self._callback_str = callback

        self.default_kwargs = kwargs
        self.name = name

        self.add_prefix(prefix)

    def match_found(
        self, path: str, match
    ) -> Optional[Tuple[Callable, Iterable[Any], Dict[str, Any]]]:
        kwargs = match.groupdict()

        if kwargs:
            args: Iterable = tuple()
        else:
            args = match.groups()

        kwargs.update(self.default_kwargs)

        return self.callback, args, kwargs

    @property
    def callback(self):
        if self._callback is not None:
            return self._callback
        try:
            self._callback = import_module(self._callback_str)
        except ImportError as e:
            raise ViewDoesNotExist(
                'Could not import %s (%s)' % (self._callback_str, str(e))
            )

        return self._callback

    def add_prefix(self, prefix):
        if prefix and hasattr(self, '_callback_str'):
            self._callback_str = prefix + '.' + self._callback_str


class RegexURLResolver(RegexURL):
    def __init__(self, regex, router, kwargs={}):
        super(RegexURLResolver, self).__init__(regex)

        if callable(router):
            self._router = router
        else:
            self._router = None
            self._router_str = router

        self.default_kwargs = kwargs

    def match_found(
        self, path: str, match
    ) -> Optional[Tuple[Callable, Iterable[Any], Dict[str, Any]]]:
        new_path = path[match.end() :]

        try:
            callback, args, kwargs = self.router.resolve(new_path)
        except Http404:
            return None

        kwargs.update(self.default_kwargs)
        return callback, args, kwargs

    @property
    def router(self):
        if self._router is not None:
            return self._router
        try:
            self._router = import_module(self._router_str)
        except ImportError as e:
            raise Http404(
                'Could not import %s. Error was: %s' % (self._router_str, str(e))
            )

        return self._router


class BaseRouter(object):
    def __init__(self, *urls):
        """
        Router takes URLs which you can register on creation.

        Example::

            router = rivr.Router(
                (r'^$', index),
                (r'^test/$', test),
            )
        """

        self.urlpatterns = []

        try:
            if isinstance(urls[0], str):
                prefix = urls[0]
                urls = urls[1:]
            else:
                prefix = None
        except IndexError:
            prefix = None

        for t in urls:
            if isinstance(t, (list, tuple)):
                t = url(*t)

            if prefix and hasattr(t, 'add_prefix'):
                t.add_prefix(prefix)

            self.urlpatterns.append(t)

    def __iadd__(self, router):
        self.urlpatterns.__iadd__(router.urlpatterns)

    def __isub__(self, router):
        self.urlpatterns.__isub__(router.urlpatterns)

    def append(self, url):
        self.urlpatterns.append(url)

    def register(self, *t):
        """
        Register a URL pattern with a view. This can either be used as a
        decorator, or it can be used as a method with a view.

        Decorator Example::

            @router.register(r'^$')
            def view(request):
                return Response()

        View Example::

            router.register(r'^$', view)

        """

        if isinstance(t, (list, tuple)):
            if len(t) == 1:

                def func(view):
                    self.register(t[0], view)
                    return view

                return func

            t = url(*t)

        self.append(t)


class Router(BaseRouter):
    append_slash = True
    """
    When append_slash is True, if the request URL does not match any patterns
    in the router and it doesn't end in a slash. The router will HTTP redirect
    any issues to the same URL with a slash appended.
    """

    def resolve(self, path: str) -> Tuple[Callable, Iterable[Any], Dict[str, Any]]:
        for pattern in self.urlpatterns:
            result = pattern.resolve(path)
            if result is not None:
                return result
        raise Resolver404('No URL pattern matched.')

    def __call__(self, request: Request) -> Response:
        if self.append_slash and (not request.path.endswith('/')):
            if (not self.is_valid_path(request.path)) and self.is_valid_path(
                request.path + '/'
            ):
                return ResponsePermanentRedirect(request.path + '/')

        result = RegexURLResolver(r'^/', self).resolve(request.path)
        if result:
            callback, args, kwargs = result
            return callback(request, *args, **kwargs)
        raise Resolver404('No URL pattern matched.')

    def is_valid_path(self, path: str) -> bool:
        try:
            self.resolve(path)
            return True
        except Resolver404:
            return False


class Domain(BaseRouter):
    APPEND_SLASH = True

    def resolve(self, path: str) -> Tuple[Callable, Iterable[Any], Dict[str, Any]]:
        for pattern in self.urlpatterns:
            result = pattern.resolve(path)
            if result is not None:
                return result
        raise Resolver404('No URL pattern matched.')

    def __call__(self, request: Request) -> Response:
        host = request.META.get('HTTP_HOST', 'localhost:80')
        host = ':'.join(host.split(':')[:-1])
        url = host + request.path

        if self.APPEND_SLASH and (not url.endswith('/')):
            if (not self.is_valid_url(url)) and self.is_valid_url(url + '/'):
                return ResponsePermanentRedirect(url + '/')

        result = self.resolve(url)
        if result:
            callback, args, kwargs = result
            return callback(request, *args, **kwargs)
        raise Resolver404('No URL pattern matched.')

    def is_valid_url(self, path: str) -> bool:
        try:
            self.resolve(path)
            return True
        except Resolver404:
            return False


# Shortcuts
def url(regex, view: Callable[..., Response], kwargs={}, name=None, prefix=None):
    if isinstance(view, list):
        return RegexURLResolver(regex, view[0], kwargs)

    as_view = getattr(view, 'as_view', None)
    if as_view and callable(as_view):
        view = as_view()

    return RegexURLPattern(regex, view, kwargs, name, prefix)


def include(router):
    return [router]
